/*

        analyze methods, routines, dependency trees, classes, class methods, instance methods, ...

        switch: if given and "1", then do not resolve (analyze) ::REQUIRES - files
                if given and "2", then return the file in question stripped off comments etc. ...
                else ::REQUIRES - directives are being followed


program:   orx_analyze.rex
type:      Object REXX, REXXSAA 6.0
purpose:   parses a REXX-program and puts the results into various variables, resolves ::REQUIRES
version:   0.9.3
date:      1995-11
changed:   1997-09-28, ---rgf, changed parsing of REQUIRES-directive, now removing quotes, also
                               find_file2() first tests existence of file before doing a find_file()
           1999-01-09, ---rgf, changed selector from "OS2ENVIRONMENT" to "ENVIRONMENT" for portability reasons
           2005-05-28, ---rgf, renamed from .cmd to .rex, uppercased filename in ::requires for ..IX-systems
           2009-06-11, ---mdl, manually load the 'class' class as .object~subclasses is broken in ORexx 2.x
           2009-08-18, ---mdl, update patch to only manually load the 'class' class when it is not already loaded

author:    Rony G. Flatscher
           Rony.Flatscher@wu-wien.ac.at
           (Wirtschaftsuniversitaet Wien, University of Economics and Business
           Administration, Vienna/Austria/Europe)
needs:     ---

usage:     orx_analyze some_rexx_file [ 1 | 2 ]

           no switch    ... analyze and populate variables
           1            ... analyze and populate variables, but do not resolve ::REQUIRES
           2            ... merely strip comments from code and break lines where ";" found,
                            returns file without comments

comments:  not finished yet

All rights reserved, copyrighted 1995-1997 by the author


*/

PARSE ARG start_file, switch

.local ~ resolve = \(switch = 1)        /* resolve ::REQUIRES, i.e. parse ::REQUIRES files ? */
.local ~ analyze.local = .directory ~ new
.local ~ analyze.local ~ selector    = "ENVIRONMENT" /* for VALUE() function */
.local ~ analyze.local ~ path        = "PATH"

IF switch = 2 THEN                      /* merely return the file stripped off comments etc. ... */
DO
   file_name = find_file2( start_file   )        /* look for file in path */
   workFile = .stream ~ new( file_name ) ~~ open("read") ~ arrayin    /* read file into array */
   RETURN     remove_comments( workFile )       /* get rid of all comments, join/split lines */
END


/* determine classes in environment to be ignored, because they are defined by this program set */
/*
.local ~ ignoreClasses = "DEF DEF_CLASS DEF_FILE DEF_LABEL DEF_METHOD DEF_PROCEDURE DEF_REQUIRES DEF_ROUTINE",
                         "ANCHOR CODEPAGE2HTML HTML_DOC HTML_LIST HTML_TABLE NLS REFERENCE RELTABLE RELUNIQUE"
*/
.local ~ ignoreClasses = ""     /* show all classes at present time     */


/* define a def_class-sentinel, indicating a missing class */
.local ~ missing.class               = .def_class ~ new( "*ERROR* N/A",,
                                                         .true, .nil, .nil, .nil, .nil, .nil )

/* build stem, containing object for cross-referencing */
ctl. = .nil
ctl.eMissingClass  = .missing.class     /* return sentinel object, indicating a missing class */
ctl.eRootObject    = .nil               /* def_file - object of Object REXX' .object */
ctl.eRootClass     = .nil               /* def_file - object of Object REXX' .class  */

/* ---->
ctl.eClasses  = .directory ~ new        /* relates classIDs to class objects */
ctl.eMethods  = .directory ~ new        /* relates methodIDs to meth objects */
ctl.eRoutines = .directory ~ new        /* relates routineIDs to routine objects */
<---- */

ctl.eFiles    = .directory ~ new        /* relates files to file objects */
ctl.eFileStart = .nil                   /* store the file which starts the analyzing */
ctl.eFileEnvironment = .nil             /* store the pseudo file containing all classes reachable from
                                           .object, not necessarily public ! */
ctl.eFilesSeq = .list      ~ new        /* store sequence of files */

ctl.eToken2Files      = .relation ~ new  /* relates all "tokens" (directives, procedures, classes, ...) to files */
ctl.eClasses2Files    = .relation ~ new  /* relates classes    to files */
ctl.eMethods2Files    = .relation ~ new  /* relates methods    to files */
ctl.eProcedures2Files = .relation ~ new  /* relates procedures to files */
ctl.eRoutines2Files   = .relation ~ new  /* relates routines   to files */
ctl.eLabels2Files     = .relation ~ new  /* relates labels     to files */
ctl.eRequires2Files   = .relation ~ new  /* relates requires   to files */


ctl.eProcedures2Objects = .relation ~ new/* relates procedures to methods, routines or files */
ctl.eLabels2Objects   = .relation ~ new  /* relates labels     to files, procedures, routines, methods */

ctl.eMethods2Class    = .relation ~ new  /* relates classes and methods */
ctl.eClassMethods2Class   = .relation ~ new  /* relates classes and *class* methods */
ctl.eInstanceMethods2Class   = .relation ~ new  /* relates classes and *instance* methods */

ctl.eEnvClassSet      = .set ~ new       /* set of all environment classes (def_class) */



/* QUITE a performance penalty                                  */
/* retrieve methods of classes defined in .environment ?        */
IF .bQueryEnvClassMethods = ".BQUERYENVCLASSMETHODS" THEN
   .local ~ bQueryEnvClassMethods      = .true
SAY ".bQueryEnvClassMethods" pp( .bQueryEnvClassMethods )

/*
trace ?i
say "go ahead, rgf"
*/


/* now do the work */

CALL build_environment_classes          /* build a pseudo-file containing .object's subclasses, */
CALL analyze_file  start_file           /* analyze file, build repositories */



RETURN ctl.

/* ------------------------------------------------------------------------------------- */

/* produce a virtual file, containing all classes presently accessible via .object (i.e. its
   subclasses; note metaclasses */

BUILD_ENVIRONMENT_CLASSES : PROCEDURE EXPOSE ctl.

   file_name = ".object-tree (from .environment)"
   CALL sayerror "Analyzing .environment" pp( file_name ) "..."
   tmpFile = .def_file ~ new( file_name )            /* create file-object */
   tmpFile ~ shortname = ".environment"              /* give it a short name too */

   ctl.eFiles ~ setentry( file_name, tmpFile )       /* remember file */
   ctl.eFileEnvironment = tmpFile                    /* pseudo file containing classes reachable from .object */

   /* the following variable gets exposed and serves as a
      directory of classes defined within this module which should get ignored */




   CALL create_defs_from_environment .object, tmpFile           /* create empty def_class'es for environment classes */
   CALL object_sub_classes .object, .false, tmpFile, .object, .set ~ new    /* setup def_class'es           */

   /* now define the metaclass for .object and .class within def_class */
   tmpEnvironmentClasses = ctl.eFileEnvironment ~ Local_Classes         /* get .object's class tree */

   tmpClassObject = tmpEnvironmentClasses ~ entry( "class"  )
   tmpClassObject ~ metaclass       = .class ~ metaclass ~ id
   tmpClassObject ~ IsMetaClass     = .true
   tmpClassObject ~ MetaClassObject = tmpClassObject
   ctl.eRootClass     = tmpClassObject
   tmpFile ~ Local_Root_Classes ~ put( tmpClassObject )

   tmpObject = tmpEnvironmentClasses ~ entry( "object"  )
   tmpObject ~ metaclass = .object ~ metaclass ~ id
   tmpObject ~ MetaClassObject = tmpClassObject
   tmpObject ~ SuperClassObject = .nil                  /* .object has *no* superclass ! */
   ctl.eRootObject     = tmpObject
   tmpFile ~ Local_Root_Classes ~ put( tmpObject )

   /* put into .local */
   .local ~ Env.Object  = tmpObject
   .local ~ Env.Class   = tmpClassObject
   .local ~ Env.FileObj = tmpFile

   .local ~ HandledClassSet = .set ~ new                /* defined for procedure/routine FIX_UP_TREE */
   CALL Fix_Up_Tree tmpFile, tmpObject, tmpObject ~ MetaClass, tmpObject ~ MetaClassObject, tmpObject ~ IsMetaClass

   RETURN




/* ------------------------------------------------------------------------------------- */
/* insert metaclass of superclass, if empty, propagate IsMetaClass value */
FIX_UP_TREE: PROCEDURE EXPOSE ctl.
   USE ARG tmpFile, aClassObj, MetaClass, MetaClassObject, IsMetaClass

   .HandledClassSet ~ put( aClassObj )                  /* memorize classes already handled */
   IF aClassObj ~ MetaClassObject = .nil THEN           /* no metaclass defined as of yet ! */
   DO
      aClassObj ~ MetaClass       = MetaClass
      aClassObj ~ MetaClassObject = MetaClassObject
      aClassObj ~ IsMetaClass     = IsMetaClass
   END
   ELSE         /* metaclass definition exists, use it */
   DO
      MetaClass       = aClassObj ~ MetaClass
      MetaClassObject = aClassObj ~ MetaClassObject
      IsMetaClass     = aClassObj ~ IsMetaClass

      IF MetaClass = .nil THEN                  /* no metaclass defined, use superclass' metaclass */
      DO
         MetaClassObject = aClassObj ~ superclassObject ~ MetaClassObject
         MetaClass = aClassObj ~ superclassObject ~ MetaClass
         IsMetaClass = aClassObj ~ superclassObject ~ IsMetaClass

         aClassObj ~ MetaClass       = MetaClass
         aClassObj ~ MetaClassObject = MetaClassObject
         aClassObj ~ IsMetaClass     = IsMetaClass
      END
   END

   IF IsMetaClass THEN
      tmpFile ~ Local_Metaclasses ~ setentry( aClassObj ~ name, aClassObj )  /* save metaclass */

   /* handling environment ? (adding to environment's metaclasses o.k. ? ) */
   bAddToSet = ( tmpFile = ctl.eFileEnvironment )

   IF \bAddToSet THEN     /* not environment: don't add metaclass usage to environment for making reading easier! */
   DO
      /* is metaclass in environment, if so, do *not* add it (for easier reading of environment) */
      bAddToSet = \ ctl.eClasses2Files ~ HasItem( MetaClassObject, ctl.eFileEnvironment )
   END

   IF bAddToSet THEN
      MetaClassObject ~ MetaUsedBySet ~ put( aClassObj )  /* save class using this specific metaclass */


   /* recurse subclasses */
   DO tmpItem OVER aClassObj ~ SetOfSubclasses
      IF \ .HandledClassSet ~ hasindex( tmpItem ) THEN           /* unhandled class in hand, recurse */
         CALL fix_up_tree tmpFile, tmpItem, MetaClass, MetaClassObject, IsMetaClass
   END

   RETURN





/* ------------------------------------------------------------------------------------- */
/* create def_class-objects for environment-classes in order to be able to reference
   disparate classes, e.g. for multiple inheritance                                     */
CREATE_DEFS_FROM_ENVIRONMENT : PROCEDURE EXPOSE ctl.
  USE ARG class, tmpFile

  /* create def_class                                                                   */
  metaclass  = .nil
  mixinclass = .nil
  subclass   = .nil
  name = class ~ ID                             /* save ID of class                     */

  /* public class, directly accessible via .environment ?                               */
  bIsPublic = ( ( "." || name ) <> TRANSLATE( "." || name ) )

  tmpObject = .def_class ~ new( "." || name, bIsPublic, .nil, .nil,  .nil, .nil, .nil )
  tmpFile ~ local_Classes ~ setentry( name, tmpObject ) /* add to environment directory */

  /* iterate over all subclasses by recursing */
  subClasses = class ~ SUBCLASSES               /* get all present subclasses           */

  /* Rony approved of my suggested patch to get Orx_Analyze to work on ORexx 2.x */
  /* Remains broken in ooRexx 3.x, working in the beta of ooRexx 4 however */
  /* Only add it if we are operating at the class root level */
  /* 20090818 Only add the Class Class if it is not already in the list */
  /* Should now work consistently with all versions of Object Rexx */
  if class~id == .object~id then
    do
      tmpClassClassID = .class~id
      tmpFoundClassClassFlag = .false
      do atemp over subClasses
        if atemp~id = tmpClassClassID then
          do
            tmpFoundClassClassFlag = .true
            leave
          end
      end

      /* Add the Class Class if not found */
      if tmpFoundClassClassFlag = .false then
        do
          tmpClassCount = subClasses~items + 1
          subClasses~put(.class, tmpClassCount)
        end
    end

  DO subclass OVER subClasses                   /* iterate over subclasses              */
     say 'CALL create_defs_from_environment  ' || subclass || ', ' || tmpFile
     CALL create_defs_from_environment  subclass, tmpFile               /* recurse      */
  END

  RETURN


/* ------------------------------------------------------------------------------------- */
/* set up all classes of .object, as if they were defined in a file */
OBJECT_SUB_CLASSES : PROCEDURE EXPOSE ctl.
  USE ARG class, IsMetaClass, tmpFile, SuperClass, HandledSet


  name = class ~ ID                                     /* save ID of class */
  tmpObject = tmpFile ~ local_classes ~ entry( name )   /* get def_class for class              */

  IF HandledSet ~ HasIndex( name ) THEN                 /* already handled                      */
     RETURN

  HandledSet ~ put( name )                              /* indicate, that class was handled     */

  metaclass  = .nil
  mixinclass = .nil
  subclass   = .nil

  name = class ~ ID                                     /* save ID of class */

  IF class <> .object THEN                              /* save superclass                      */
  DO
     IF class ~ QUERYMIXINCLASS THEN                    /* is class a mixinclass ?              */
        mixinclass = SuperClass ~ ID                    /* save ID of class                     */
     ELSE
        subclass = SuperClass ~ ID                      /* save ID of class                     */
  END

  IF class ~ METACLASS <> .class THEN                   /* metaclass, other than .class ?       */
  DO
     metaclass = class ~ METACLASS ~ ID                 /* save ID of metaclass                 */
     IF class ~ METACLASS = class THEN                  /* a metaclass itself ?                 */
        IsMetaClass = .true                             /* set metaclass indicator on           */
  END


  IF .bQueryEnvClassMethods THEN
  DO
   /* */
     tmpMethodSupp = class ~ METHODS( .nil )            /* get instance methods */
     tmpMethDir = .directory ~ new
     DO WHILE tmpMethodSupp ~ available
        tmpMethString = tmpMethodSupp ~ index           /* get method's name */
        tmpMethObj = .def_method ~ new( tmpMethString, .false, .false, .false, .false, .false )    /* create a def_meth for it */
        tmpMethDir ~ setentry( tmpMethString, tmpMethObj ) /* save into directory */
        tmpMethodSupp ~ next
     END
     bClassMethod = .false
   /* */

     /* if a class' metaclass is the class itself, then the class is a metaclass ! */
     IF class ~ METACLASS = class THEN
     DO
        bClassMethod = .true                               /* use instance methods as class methods too */
        tmpMethClassDir = .directory ~ new                 /* copy methods */
        DO item OVER tmpMethDir
           newMeth = .def_method ~ new( item, .false, .false, .false, .false, .true )    /* create a def_meth for it */
           tmpMethClassDir ~ setentry( item, newMeth )
        END

   /* */
     END
  END


  /* get def_class for class at hand */
  tmpObject = tmpFile ~ local_classes ~ entry( name )           /* get def_class                */
  tmpObject ~ subclass   = subclass                             /* set value                    */
  tmpObject ~ mixinclass = mixinclass                           /* set value                    */
  tmpObject ~ metaclass  = metaclass                            /* set value                    */


  IF .bQueryEnvClassMethods THEN
  DO
   /* */
     tmpObject ~ Local_Instance_Methods = tmpMethDir       /* save instance methods */

     IF bClassMethod THEN
        tmpObject ~ Local_Class_Methods = tmpMethClassDir  /* save as class methods too */
   /* */
  END

  tmpObject ~ IsMetaClass = IsMetaClass                 /* set metaclass indicator with class */
  tmpObject ~ LineNr = .nil                             /* indicate, no line # available */
  ctl.eEnvClassSet ~ put( tmpObject )                   /* save environment class in set */

  tmpObject ~ SuperClassObject = tmpFile ~ local_Classes ~ entry( SuperClass ~ id )


  ctl.eToken2Files[ tmpFile ]   = tmpObject             /* add to relation */
  ctl.eClasses2Files[ tmpFile ] = tmpObject             /* add to relation */

  IF isMetaClass THEN                                           /* store metaclass with file */
  DO
     /* root of a new, self-contained metaclass found ? */
     IF class ~ id = tmpObject ~ METACLASS THEN
        tmpFile ~ Local_Root_Classes ~ put( tmpObject )
  END

  IF metaclass <> .nil THEN                             /* metaclass given, if so get def_object */
     tmpObject ~ metaClassObject = tmpFile ~ local_Classes ~ entry( metaclass )



  /* set superclasses in def_object style */
  tmpLocalClasses = tmpFile ~ local_Classes     /* get local-classes directory */
  tmpArray = class ~ SUPERCLASSES               /* get superclasses */

  IF tmpArray ~ items = 0 THEN                  /* no superclasses, hence root */
     tmpFile ~ Local_Root_Classes ~ put( tmpObject )
  ELSE
  DO
     DO i = 1 TO tmpArray ~ items
        tmpObject ~ ListOfSuperClasses ~ insert( tmpLocalClasses ~ entry( tmpArray[ i ] ~ id ) )
     END
  END


  /* iterate over all subclasses by recursing */
  subClasses = class ~ SUBCLASSES               /* get all present subclasses           */

  /* Rony approved of my suggested patch to get Orx_Analyze to work on ORexx 2.x */
  /* Remains broken in ooRexx 3.x, working in the beta of ooRexx 4 however */
  /* Only add it if we are operating at the class root level */
  /* 20090818 Only add the Class Class if it is not already in the list */
  /* Should now work consistently with all versions of Object Rexx */
  if class~id == .object~id then
    do
      tmpClassClassID = .class~id
      tmpFoundClassClassFlag = .false
      do atemp over subClasses
        if atemp~id = tmpClassClassID then
          do
            tmpFoundClassClassFlag = .true
            leave
          end
      end

      /* Add the Class Class if not found */
      if tmpFoundClassClassFlag = .false then
        do
          tmpClassCount = subClasses~items + 1
          subClasses~put(.class, tmpClassCount)
        end
    end

  DO subclass OVER subClasses                   /* iterate over subclasses              */
     CALL object_sub_classes subclass, IsMetaClass, tmpFile, class, HandledSet      /* recurse      */
  END

  /* now that all subclasses have been defined, set subclasses in def_object style */
  tmpArray = class ~ SUBCLASSES                 /* get subclasses */

  IF tmpArray ~ items = 0 THEN                  /* no subclasses, hence leaf */
     tmpFile ~ Local_Leaf_Classes ~ put( tmpObject )
  ELSE
  DO
     DO i = 1 TO tmpArray ~ items
        tmpObject ~ SetOfSubclasses ~ put( tmpLocalClasses ~ entry( tmpArray[ i ] ~ id ) )
     END
     tmpObject ~ IsUsedAsSuper = .true          /* class is used as a superclass ! */
  END

  RETURN







/* ------------------------------------------------------------------------------------- */
/* fix-up all class pointers, especially mark metaclasses and its descendants as such */
FIX_UP_CLASSES: PROCEDURE EXPOSE ctl.
   USE ARG tmpFile, wfLength

   tmpVisibleClasses     = tmpFile ~ Local_Classes ~ union( tmpFile ~ Visible_Classes ) /* get visible classes */
   tmpEnvironmentClasses = ctl.eFileEnvironment ~ Local_Classes         /* get .object's class tree */

   tmpSupp  = tmpFile ~ local_Classes ~ supplier        /* get a supplier from directory */

   DO WHILE tmpSupp ~ available                         /* build list of superclasses and set of subclasses */
      supp_Item  = tmpSupp ~ Item                       /* get def_class         */
      supp_Index = tmpSupp ~ Index                      /* get name of def_class */

      tmpIsMetaClass    = .false                        /* default */
      line?             = supp_item ~ LineNr            /* get line # of class-definition */

      tmpSuperClassString = supp_Item ~ subclass        /* get superclass classid */

      IF tmpSuperClassString = .nil THEN
         tmpSuperClassString = supp_Item ~ mixinclass

      IF tmpSuperClassString <> .nil THEN               /* find class in present's file visible classes */
      DO
         tmpSuperClassStringMixCase = tmpSuperClassString
         tmpSuperClassString = TRANSLATE( tmpSuperClassString )    /* translate to uppercase */
         tmpSuperClass = tmpVisibleClasses ~ entry( tmpSuperClassString )  /* is the class defined ? */
         IsHandled = .true                      /* default to handled */

         IF tmpSuperClass = .nil THEN           /* class not found so far, check whether it is CLASS or OBJECT */
         DO
            IsHandled = .false                  /* local root, if metaclass, fix-up subclasses, found so far */

            tmpSuperClass = tmpEnvironmentClasses ~ entry( tmpSuperClassString )  /* is the class defined ? */
            IF tmpSuperClass <> .nil THEN       /* a class was found ! */
            DO
               tmpIsMetaclass = tmpSuperClass ~ IsMetaClass  /* store whether it is a metaclass or not */
               tmpFile ~ Local_Root_Classes ~ put( supp_item )

               /* in .object-tree ! */
               supp_item ~ SuperClassObject = tmpSuperClass  /* save superclass' def_object */
               supp_item ~ ListOfSuperClasses ~ insert( tmpSuperClass )    /* add superclass def_object */
            END
            ELSE                /* error, missing class ! */
            DO
               tmpString = pp_lineNr( line?, wfLength ) pp( "Unknown Class" ) "- class" pp( supp_item ~ name),
                           "attempts to subclass/mixinclass unknown class:" pp( tmpSuperClassStringMixCase ) "!"
               tmpFile ~ errors ~ insert( tmpString)                    /* save error with file */
               supp_item ~ errors ~ insert( tmpString )                 /* save error with class */
            END
         END
         ELSE                           /* add subclasses to visible class */
         DO
            tmpSuperClass ~ IsUsedAsSuper = .true                       /* class is used as a superclass ! */
            tmpSuperClass ~ SetOfSubclasses ~ put( supp_item )          /* put child into parents set */
            supp_item ~ ListOfSuperClasses ~ insert( tmpSuperClass )    /* add superclass def_object */
            supp_item ~ IsMetaClass = tmpSuperClass ~ IsMetaClass       /* child is of same type as parent */
            supp_item ~ SuperClassObject = tmpSuperClass                /* save superclass' def_object */
         END

         /* ok actual class is a metaclass and so are all classes referring to this one */
         IF tmpIsMetaclass THEN
         DO
            /* add to metaclass directory */
            tmpFile ~ Local_Metaclasses ~ setentry( supp_item ~ name, supp_item )

            IF \IsHandled THEN
               CALL fix_up_metaclasses supp_item, tmpFile       /* fix up all descendants (they are metaclasses too!) */
         END
      END


      tmpInherit = supp_Item ~ inherit                  /* handle all inherited superclasses */
      IF tmpInherit <> .nil THEN                        /* mark all inherited classes as superclasses */
      DO
         DO inh_Item OVER supp_Item ~ inherit           /* mark all inherited classes as superclasses */
            tmpSuperClass = tmpVisibleClasses ~ entry( inh_Item )       /* is the class defined ? */

            IF tmpSuperClass <> .nil THEN
            DO
               tmpSuperClass ~ IsUsedAsSuper = .true                    /* class is used as a superclass ! */
               tmpSuperClass ~ SetOfSubclasses ~ put( supp_item )       /* put child into parents set */
               supp_item ~ ListOfSuperClasses ~ insert( tmpSuperClass ) /* add superclass def_object */
            END
            ELSE        /* class is missing, is it in the environment ?  */
            DO
               tmpSuperClass = tmpEnvironmentClasses ~ entry( inh_item )/* is the class defined ? */

               IF tmpSuperClass = .nil THEN
               DO
                  tmpString = pp_lineNr( line?, wfLength ) pp( "Unknown Class" ) "- class" pp( supp_item ~ name) "attempts to inherit unknown class:" pp( inh_item ) "!"
                  tmpFile ~ errors ~ insert( tmpString)                 /* save error with file */
                  supp_item ~ errors ~ insert( tmpString )              /* save error with class */
               END
               ELSE     /* add found class list of superclasses */
               DO
                  supp_item ~ ListOfSuperClasses ~ insert( tmpSuperClass ) /* add superclass def_object */
               END
            END
         END
      END

      /* check whether a referred to metaclass exists */
      tmpMetaClass = supp_Item ~ metaclass
      IF tmpMetaClass <> .nil THEN
      DO
         tmpMetaClassObject = tmpVisibleClasses ~ entry( tmpMetaClass ) /* try to get metaclass object */

         IF tmpMetaClassObject = .nil THEN      /* error, class is missing, try to find it in environment  */
            tmpMetaClassObject = tmpEnvironmentClasses ~ entry( inh_item  )

         IF tmpMetaClassObject = .nil THEN      /* metaclass not found, really ! */
         DO
            tmpString = pp_lineNr( line?, wfLength ) pp( "Unknown Class" ) "- class" pp( supp_item ~ name),
                        "attempts to use unknown metaclass:" pp( tmpMetaClass ) "!"
            tmpFile ~ errors ~ insert( tmpString)                 /* save error with file */
            supp_item ~ errors ~ insert( tmpString )              /* save error with class */
         END
         ELSE
         DO
            supp_item ~ MetaClassObject = tmpMetaClassObject    /* save metaclass object */
            tmpMetaClassObject ~ MetaUsedBySet ~ put( supp_item)/* save class within metaclass */
         END
      END

      tmpSupp ~ next                    /* get next class */
   END


   /* determine local root- and leaf-classes, fill in environment's .object or .class as appropriate */
   tmpSupp  = tmpFile ~ local_Classes ~ supplier        /* get a supplier from directory */

   envObjectObject = tmpEnvironmentClasses ~ entry( "object" )  /* get def_class */
   envClassObject  = tmpEnvironmentClasses ~ entry( "class" )   /* get def_class */

   /* determine local root classes and leaf classes, check whether referred metaclasses are in effect metaclasses ... */
   DO WHILE tmpSupp ~ available
      supp_Item  = tmpSupp ~ Item
      supp_Index = tmpSupp ~ Index

      IF supp_item ~ ListOfSuperClasses ~ items = 0 THEN     /* a local root class, possibly going into .object's subclasses */
      DO
         tmpFile ~ Local_Root_Classes ~ put( supp_item )

         IF supp_item ~ IsMetaClass THEN
         DO
            supp_item ~ ListOfSuperClasses ~ insert( envClassObject, .nil ) /* add superclass def_object */
            supp_item ~ SuperClassObject = envClassObject                   /* save superclass' def_object */
         END
         ELSE
         DO
            supp_item ~ ListOfSuperClasses ~ insert( envObjectObject, .nil ) /* add superclass def_object */
            supp_item ~ SuperClassObject = envObjectObject                   /* save superclass' def_object */

            /*
            supp_item ~ listOfSuperClasses ~ insert( envObjectObject )       /* insert def_object into list of superclasses */
            */
         END
      END

      IF supp_item ~ SetOfSubclasses ~ items = 0 THEN
         tmpFile ~ Local_Leaf_Classes ~ put( supp_item )

      /* Referred to metaclass not a metaclass ? */
      tmpObject = supp_item ~ metaClassObject                   /* get referred to metaclass object */
      IF tmpObject <> .nil & tmpObject <> .missing.class THEN
      DO
         IF \ tmpObject ~ IsMetaClass THEN
         DO
            tmpString = pp_lineNr( line?, wfLength ) pp( "Not a Metaclass" ) "- class" pp( supp_item ~ name),
                        "attempts to use object class" pp( tmpObject ~ name ) "as a metaclass !"
            tmpFile ~ errors ~ insert( tmpString)               /* save error with file */
            supp_item ~ errors ~ insert( tmpString )            /* save error with class */
            tmpObject ~ MetaUsedBySet = .set ~ new              /* make sure set is empty */
            tmpFile ~ Local_Metaclasses ~ remove( tmpObject ~ name )    /* remove from metaclass directory */
         END
      END

      tmpSupp ~ next                    /* get next class */
   END


   /* fix up tree for every local root */
   DO aClass OVER tmpFile ~ Local_Root_Classes
      /* if metaclass is missing in a local root, add superclass' metaclass */
      IF aClass ~ MetaClass = .nil THEN
      DO
         aSuper = aClass ~ SuperClassObject
         aClass ~ MetaClass       = aSuper ~ MetaClass
         aClass ~ MetaClassObject = aSuper ~ MetaClassObject
         aClass ~ IsMetaClass     = aSuper ~ IsMetaClass
      END
      CALL Fix_Up_Tree tmpFile, aClass, aClass ~ MetaClass, aClass ~ MetaClassObject, aClass ~ IsMetaClass
   END

   RETURN



/* ------------------------------------------------------------------------------------- */
/* all subclasses are metaclasses too ! */
FIX_UP_METACLASSES: PROCEDURE
   USE ARG aClass, tmpFile

   aClass ~ IsMetaClass = .true                         /* mark as metaclass */

   tmpSupp = aClass ~ SetOfSubclasses ~ supplier        /* get all presently defined subclasses */
   DO WHILE tmpSupp ~ available
       CALL fix_up_metaclasses tmpSupp ~ item, tmpFile  /* work on this class */
       tmpSupp ~ next
   END
   RETURN




/* ------------------------------------------------------------------------------------- */

/* read file, analyze it, resolve ::REQUIRES directive ... */
ANALYZE_FILE: PROCEDURE EXPOSE ctl.
   USE ARG   file_name_in, called_from_file


   IF \VAR( "called_from_file" ) THEN called_from_file = .nil

   file_name = find_file2( file_name_in )        /* look for file in path */

   analyzed = ctl.eFiles ~ entry(file_name )    /* check, whether already analyzed */

   IF  analyzed <> .nil THEN
   DO
      IF called_from_file <> .nil THEN
      DO
         called_from_file ~ requires_files ~ insert( analyzed  )         /* save with file containing ::requires */
         analyzed         ~ required_by    ~ insert( called_from_file )  /* save file, which requires this one   */
      END

      CALL sayError pp( file_name ) || ": file has been analyzed already ..."
      RETURN analyzed                           /* return file-object */
   END

   tmpFile = .def_file ~ new( file_name )            /* create file-object */

   IF called_from_file <> .nil THEN                   /* save calling file */
   DO
      called_from_file ~ requires_files ~ insert( tmpFile )         /* save with file containing ::requires */
      tmpFile          ~ required_by    ~ insert( called_from_file )/* save file, which requires this one   */
   END
   ELSE
      ctl.eFileStart = tmpFile                       /* this is the file, which trickles analyzing ... */


   ctl.eFiles ~ setentry( file_name, tmpFile )       /* remember file */
   ctl.eFilesSeq ~ INSERT( tmpFile )


   workFile = .stream ~ new( file_name ) ~~ open("read") ~ arrayin    /* read file into array */
   wfLength = LENGTH( workFile ~ items )        /* length of largest line number */
   tmpFile ~ total_loc = workFile ~ items       /* save lines of code */

/*
say "remove_comments from" pp( file_name ) "begin ..."
*/
   newArray = remove_comments( workFile )       /* get rid of all comments, join/split lines */
/*
say "remove_comments from" pp( file_name ) "END."
*/

   newArrayItems = newArray ~ items / newArray ~ dimension
   tmpFile ~ loc = newArrayItems                /* save lines of code (stripped) */

   CALL sayError pp( file_name) "analyzing ..."

   tmpClass  = .nil                     /* no classes defined as of yet */
   tmpMethod = .nil                     /* no methods defined as of yet */
   tmpObject = .nil                     /* no object created so far */
   tmpProcedure = .nil                  /* no procedure object created so far */
   tmpRoutine   = .nil                  /* no routine object created so far */

   inDirective = .false
   noRequires = .false       /* a directive other than ::REQUIRES already processed ? */

   DO i = 1 TO newArrayItems
      line  = newArray[i, 1]            /* get line */
      line? = newArray[i, 2]

      tmpRes = get_next_token( line, 1, line? ) /* get first token                      */
      firstWord  = TRANSLATE( tmpRes ~ at( 1 ) )
      secondWord = ""

      IF tmpRes ~ at( 3 ) > 0 THEN      /* maybe tokens left                            */
      DO
         tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token    */
         secondWord = TRANSLATE( tmpRes ~ at( 1 ) )

         IF firstWord = ":" & secondWord = ":" THEN        /* directive in hand ?       */
         DO
            firstWord  = "::"
            secondWord = ""
            IF tmpRes ~ at( 3 ) > 0 THEN        /* maybe tokens left                    */
            DO
               tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token       */
               secondWord = TRANSLATE( tmpRes ~ at( 1 ) )
            END
         END
      END

/***********>
      PARSE UPPER VAR line firstWord secondWord . /* get first word as uppercase */
<***********/

/***********>
<***********/
      tmpErrors = .list ~ new                   /* create temporary list for errors to be attached to def-objects */

      SELECT
         WHEN firstWord = "USE"   THEN          /* USE ARG in hand ? */
              DO
                 IF secondWord = "ARG" THEN
                 DO
/***********>
                    PARSE VAR line . . newL     /* get argument(s) */
<***********/
                    newL = ""
                    IF tmpRes ~ at( 3 ) > 0 THEN        /* tokens left ?        */
                    DO
                       newL = STRIP( SUBSTR( line, tmpRes ~ at( 3 ) ) )
                    END
                    newL = pp_arg_list( "USE ARG" newL )

                    SELECT
                       WHEN \noRequires THEN    /* no directvies as of yet */
                            DO
                               IF tmpProcedure = .nil THEN
                                  tmpTarget = tmpFile
                               ELSE                             /* belongs to a procedure of file */
                               DO
                                  IF tmpLabel <> .nil THEN
                                     tmpTarget = tmpLabel
                                  ELSE
                                     tmpTarget = tmpProcedure
                               END
                            END

                       WHEN tmpLabel   <> .nil THEN             /* belongs to a label (terminates procedure) */
                            tmpTarget = tmpLabel

                       WHEN tmpProcedure <> .nil THEN           /* belongs to a procedure */
                            tmpTarget = tmpProcedure

                       WHEN tmpRoutine <> .nil THEN             /* belongs to a routine */
                            tmpTarget = tmpRoutine

                       WHEN tmpMethod <> .nil THEN              /* belongs to a method */
                            tmpTarget = tmpMethod

                       OTHERWISE
                            DO
                               errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 5) || "..." ),
                                    pp( "USE ARG" ) "- not defined for a routine/procedure/method !"
                               CALL sayError  errString
                               tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                            END
                            ITERATE
                    END  /* select */

                    tmpTarget ~ signatures ~ setentry( newL, newL )
                    tmpTarget ~ IsProcedure = .true


                    ITERATE                                     /* read next line */

                 END
                 ELSE
                 DO
                   errString =  pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                        "- USE *not* followed by 'ARG' !"
                   CALL sayError  errString
                   tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                 END

                 ITERATE
              END

         WHEN firstWord = "PARSE" THEN          /* PARSE ARG in hand ? */
              DO
/***********>
                 PARSE UPPER VAR line . . thirdWord fourthWord .
<***********/
                 thirdWord  = ""
                 fourthWord = ""
                 thirdPos  = tmpRes ~ at( 3 )   /* position to start parsing for third word     */
                 fourthPos = 0
                 IF tmpRes ~ at( 3 ) > 0 THEN   /* maybe tokens left                            */
                 DO
                    tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token    */
                    thirdWord = TRANSLATE( tmpRes ~ at( 1 ) )

                    fourthPos = tmpRes ~ at( 3 )
                    IF tmpRes ~ at( 3 ) > 0 THEN        /* maybe tokens left                    */
                    DO
                       tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token       */
                       fourthWord = TRANSLATE( tmpRes ~ at( 1 ) )
                    END
                 END

                 SELECT
                    WHEN secondWord = "ARG" THEN
                         DO
/***********>
                            PARSE VAR line . . newL     /* get argument(s) */
<***********/
                            newL = ""
                            IF thirdPos > 0 THEN        /* tokens left ?        */
                            DO
                               newL = STRIP( SUBSTR( line, thirdPos ) )
                            END
                            newL = pp_arg_list( "PARSE ARG" newL )
                         END

                    WHEN secondWord = "UPPER" | secondWord = "LOWER" | secondWord = "CASELESS" THEN
                         DO
                            IF thirdWord = "ARG" THEN
                            DO
/***********>
                               PARSE VAR line . . . newL        /* get arguments */
<***********/
                            newL = ""
                            IF fourthPos > 0 THEN        /* tokens left ?        */
                            DO
                               newL = STRIP( SUBSTR( line, fourthPos ) )
                            END
                               newL = pp_arg_list( "PARSE" secondWord thirdWord newL )
                            END
                            ELSE IF thirdWord = "CASELESS" & fourthWord = "ARG" THEN
                            DO
/***********>
                               PARSE VAR line . . . . newL      /* get arguments */
<***********/
                               newL = ""
                               IF tmpRes ~ at( 3 ) > 0 THEN     /* tokens left ?        */
                               DO
                                  newL = STRIP( SUBSTR( line, tmpRes ~ at( 3 ) ) )
                               END
                               newL = pp_arg_list( "PARSE" secondWord thirdWord fourthWord newL )
                            END
                            ELSE
                               ITERATE          /* not a PARSE ... ARG */
                         END

                    OTHERWISE ITERATE           /* not a PARSE ... ARG */
                 END  /* select */

                 SELECT
                    WHEN \noRequires THEN    /* no directvies as of yet */
                         DO
                            IF tmpProcedure = .nil THEN      /* belongs to file, i.e. file *is* a procedure */
                               tmpTarget = tmpFile
                            ELSE                             /* belongs to a procedure of file */
                            DO
                               IF tmpLabel <> .nil THEN
                                  tmpTarget = tmpLabel
                               ELSE
                                  tmpTarget = tmpProcedure
                            END
                         END

                    WHEN tmpLabel   <> .nil THEN             /* belongs to a label (terminates procedure) */
                         tmpTarget = tmpLabel

                    WHEN tmpProcedure <> .nil THEN           /* belongs to a procedure */
                         tmpTarget = tmpProcedure

                    WHEN tmpRoutine <> .nil THEN             /* belongs to a routine */
                         tmpTarget = tmpRoutine

                    WHEN tmpMethod <> .nil THEN              /* belongs to a method */
                         tmpTarget = tmpMethod

                    OTHERWISE
                         DO
                            errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                                 pp( "PARSE ARG" ) "- not defined for a routine/procedure/method !"
                            CALL sayError  errString
                            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                         END
                         ITERATE
                 END  /* select */

                 tmpTarget ~ signatures ~ setentry( newL, newL )    /* label may be used as a procedure */
                 tmpTarget ~ IsProcedure = .true

                 ITERATE                        /* get next line */
              END

/*
         WHEN LEFT( firstWord, 2 ) = "::" THEN  /* directive ? gets processed later on */
*/
         WHEN firstWord = "::" THEN             /* directive ? gets processed later on */
              DO
/***********>
                 PARSE VAR line "::" directive name newL        /* look for directive   */
<***********/
                 directive = secondWord                         /* get directive        */

                 name  = ""
                 IF tmpRes ~ at( 3 ) > 0 THEN                   /* get name             */
                 DO
                    tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )    /* get next token */
                    name = tmpRes ~ at( 1 )

                    newL = ""
                    IF tmpRes ~ at( 3 ) > 0 THEN                /* maybe tokens left    */
                       newL = STRIP( SUBSTR( line, tmpRes ~ at( 3 ) ) )
                 END

                 IF directive = "" | name = "" THEN
                 DO
                    errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." )

                    IF name = "" THEN errString = errString "- name of directive is missing !"
                                 ELSE errString = errString "- directive is missing !"
                    CALL sayError  errString
                    tmpFile ~ errors ~ insert( errString )   /* store error with file-object */

                    ITERATE
                 END

                 tmpMethod    = .nil                    /* reset method object */
                 tmpProcedure = .nil                    /* reset procedure object */
                 tmpRoutine   = .nil                    /* reset routine object */
                 tmpLabel     = .nil                    /* reset label object */
              END


         WHEN secondWord = ":" THEN                     /* procedure or label in hand ?         */
              DO
                 name = firstWord                       /* get name of label/procedure          */
                 next_word = ""
                 nextPos   = tmpRes ~ at( 3 )
                 IF tmpRes ~ at( 3 ) > 0 THEN           /* maybe tokens left                    */
                 DO
                    tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token       */
                    next_word = TRANSLATE( tmpRes ~ at( 1 ) )
                 END

                 procedure = (next_word = "PROCEDURE" ) /* is it a procedure or a plain label ? */
/*
say "--- 1 - procedure:" pp( procedure) "==>" pp( line )
*/

                 IF \ procedure                   THEN  /* check whether "PROCEDURE" is on next line */
                 DO
                    IF next_word = "" THEN              /* nothing left, hence look in next line */
                    DO
                       i = i + 1
                       IF i > newArrayItems THEN        /* at end already ? */
                       DO
                          i = i - 1
                       END
                       ELSE                             /* check for PROCEDURE on next line */
                       DO
                          procedure = ( TRANSLATE( SUBWORD( newArray[i, 1], 1, 1) ) = "PROCEDURE" )

                          IF procedure THEN             /* now get the position of tokens to follow "PROCEDURE" */
                          DO
                             line  = newArray[i, 1]     /* get line */
                             line? = newArray[i, 2]     /* get line# in original file */
                             tmpRes = get_next_token( line, 1, line? )  /* get the position to parse further */
                          END
                          ELSE                          /* no, no PROCEDURE statement in next line */
                          DO
                             i = i - 1                  /* reset index  */
                          END
                       END
/*
say "--- 1a- procedure:" pp( procedure) "==>" pp( line )
*/
                    END
                    ELSE                                /* a statement following label, remove label */
                    DO
                       newArray[ i, 1 ] = STRIP( SUBSTR( line, nextPos ) )      /* remove label */
/*
say "--- 1b- leaving line alone          ==>" pp( newArray[ i, 1] )
*/
                       i = i - 1                        /* have this line reread */
                    END
                 END
/*
say "--- 2 - procedure:" pp( procedure) "==>" pp( line )
*/




                 /* if a label, check whether next line makes it a procedure ! */
                 IF \ procedure THEN                    /* label found, check whether it is in fact a procedure */
                 DO
                    IF tmpClass <> .nil THEN    /* no labels within a class definition allowed ! */
                    DO
                       IF tmpMethod = .nil THEN       /* in effect, only a class method available ! */
                       DO
                          errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                               pp( "LABEL" ) "- not allowed within a CLASS-directive ! "
                          CALL sayError  errString
                          tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                          tmpClass~ errors ~ insert( errString )   /* store error with class-object */
                          ITERATE                         /* read next line */
                       END
                    END

                    name = STRIP( name )
                    tmpObject = .def_label ~ new( name, .false )
                    tmpObject ~ LineNr = line?         /* store line # of definition */
                    tmpObject ~ Errors = tmpErrors     /* assign temporary error-string */

                    /* a true label may be given within files, procedures, routines, methods */
                    SELECT                                         /* labels */

                       WHEN \noRequires THEN    /* no directives as of yet */
                            DO
                               IF tmpProcedure = .nil THEN      /* attach label to file         */
                                  tmpTarget = tmpFile
                               ELSE                             /* belongs to a procedure !     */
                               DO
                                  IF tmpProcedure <> .nil THEN  /* attach label to procedure    */
                                     tmpTarget = tmpProcedure
                                  ELSE
                                     tmpTarget = tmpLabel
                               END
                            END

                       WHEN tmpProcedure <> .nil THEN tmpTarget = tmpProcedure

                       WHEN tmpMethod <> .nil    THEN tmpTarget = tmpMethod

                       WHEN tmpRoutine <> .nil   THEN tmpTarget = tmpRoutine

                       OTHERWISE
                       DO
                          tmpTarget = .nil         /* delete invalid target */

                          errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                               pp( "LABEL" ) "- not allowed within a CLASS-directive ! "
                          CALL sayError  errString
                          tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                          tmpClass~ errors ~ insert( errString )   /* store error with class-object */
                       END
                    END  /* select */

                    /* label already defined for target object ? */
                    IF tmpTarget ~ local_Labels ~ entry( name ) <> .nil THEN
                    DO
                       errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                            pp( "LABEL" ) "- already defined for this" tmpTarget ~ type "!"
                       CALL sayError  errString
                       tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                       tmpObject ~ Errors ~ insert( errString ) /* store error with duplicate object ! */

                       IF tmpTarget <> tmpFile THEN
                          tmpTarget ~ Errors ~ insert( errString ) /* store error with target ! */
                    END
                    ELSE   /* store label with target object's directory */
                    DO
                       tmpTarget ~ local_Labels ~ setentry( name, tmpObject )
                    END

                    ctl.eLabels2Files[ tmpFile ] = tmpObject         /* add to relation */
                    ctl.eToken2Files[ tmpFile ] = tmpObject          /* add to relation */
                    ctl.eLabels2Objects[ tmpTarget    ] = tmpObject

                    tmpLabel = tmpObject           /* remember procedure object in hand */

                    ITERATE                        /* read next line */
                 END    /* IF \procedure */


                 /* o.k. a PROCEDURE in hand !  */
                 expose = ""
                 IF tmpRes ~ at( 3 ) > 0 THEN        /* get "expose" keyword                 */
                 DO
                    tmpRes = get_next_token( line, tmpRes ~ at( 3 ), line? )  /* get next token       */
                    expose = TRANSLATE( tmpRes ~ at( 1 ) )
                 END

                 newL = ""
                 IF tmpRes ~ at( 3 ) > 0 THEN        /* get expose list                      */
                 DO
                    newL = STRIP( SUBSTR( line, tmpRes ~ at( 3 ) ) )
                 END


                 IF tmpClass <> .nil THEN    /* no labels within a class definition allowed ! */
                 DO
                    IF tmpMethod = .nil THEN       /* in effect, only a class method available ! */
                    DO
                       errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                            pp( "PROCEDURE" ) "- not allowed within a CLASS-directive ! "
                       CALL sayError  errString
                       tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                       tmpClass~ errors ~ insert( errString )   /* store error with class-object */
                       ITERATE                         /* read next line */
                    END
                 END

                 tmpObject = .def_procedure ~ new( STRIP( name ), .true )
                 tmpObject ~ LineNr = line?         /* store line # of definition */
                 tmpObject ~ Errors = tmpErrors     /* assign temporary error-string */

                 IF TRANSLATE( expose ) = "EXPOSE" THEN      /* expose list there ? */
                 DO
                    tmpDir = parse_expose( newL )    /* parse expose list into a directory */
                    tmpSupp = tmpDir ~ supplier
                    DO WHILE tmpSupp ~ available     /* iterate over tmpDir, set method and class */
                       word1 = tmpSupp ~ index
                       word2 = tmpSupp ~ item
                       tmpObject ~ expose ~ setentry( word1, word2 )  /* store with procedure */
                       tmpSupp ~ next
                    END
                 END /* firstWord */


                 /* a true procedure may be given within files, procedures, routines, methods */
                 SELECT                                         /* labels */
                    WHEN tmpMethod <> .nil    THEN tmpTarget = tmpMethod

                    WHEN tmpRoutine <> .nil   THEN tmpTarget = tmpRoutine

                    OTHERWISE                      tmpTarget = tmpFile
                 END  /* select */


                 /* label already defined for target object ? */
                 IF tmpTarget ~ local_Procedures ~ entry( name ) <> .nil THEN
                 DO
                    errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                         pp( "PROCEDURE" ) "- already defined for this" tmpTarget ~ type "!"
                    CALL sayError  errString
                    tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                    tmpObject ~ Errors ~ insert( errString ) /* store error with duplicate object ! */
                    tmpTarget ~ Errors ~ insert( errString ) /* store error with target ! */
                 END
                 ELSE   /* store label with target object's directory */
                    tmpTarget ~ local_Procedures ~ setentry( name, tmpObject )


                 ctl.eToken2Files[ tmpFile ] = tmpObject             /* add to relation */
                 ctl.eProcedures2Files[ tmpFile ] = tmpObject        /* add to relation */
                 ctl.eProcedures2Objects[ tmpTarget    ] = tmpObject    /* assign procedure to def-object */

                 tmpProcedure = tmpObject       /* remember procedure object in hand */
                 tmpLabel = .nil                /* reset tmpLabel */
                 ITERATE                        /* read next line */
              END


         OTHERWISE                              /* check for statements, which transfer control */
         DO
            Keyword.1 = "RETURN"                /* keywords to look for */
            Keyword.2 = "EXIT"
            Keyword.3 = "SIGNAL"
            Keyword.4 = "RAISE"
            Keyword.0 = 4
            DO j = 1 TO Keyword.0
               tmpResult = parse_keyword( keyword.j, line ) /* check whether RETURN-keyword available */

               IF tmpResult <> .nil THEN                /* an expression was found with keyword */
               DO
                  newL = STRIP( keyword.j tmpResult )   /* precede expression with keyword */
                  LEAVE
               END

            END
            IF tmpResult = .nil THEN ITERATE    /* keyword itself not found */

            tmpTarget = .nil
            SELECT
               WHEN \noRequires THEN    /* no directvies as of yet */
                    DO
                       IF tmpProcedure = .nil THEN      /* belongs to file, i.e. file *is* a procedure */
                          tmpTarget = tmpFile
                       ELSE IF tmpLabel <> .nil THEN
                          tmpTarget = tmpLabel
                       ELSE IF tmpProcedure <> .nil THEN/* belongs to a procedure of file */
                          tmpTarget = tmpProcedure
                    END

               WHEN tmpLabel     <> .nil THEN tmpTarget = tmpLabel

               WHEN tmpProcedure <> .nil THEN tmpTarget = tmpProcedure

               WHEN tmpRoutine   <> .nil THEN tmpTarget = tmpRoutine

               WHEN tmpMethod    <> .nil THEN tmpTarget = tmpMethod

               OTHERWISE NOP
            END  /* select */


            IF tmpTarget = .nil THEN                            /* no target found, error ! */
            DO
               errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                    pp( keyword.j ) "- not defined for a routine/procedure/method/label !"
               CALL sayError  errString
               tmpFile ~ errors ~ insert( errString )           /* store error with file-object */
            END
            ELSE                                                /* add to target's return directory */
            DO
               IF keyword.j = "RETURN" THEN
               DO
                  tmpTarget ~ returns ~ setentry( newL, newL )
                  IF tmpResult <> "" THEN                       /* this object serves as a fucntion */
                     tmpTarget ~ IsFunction = .true
               END
               ELSE
                  tmpTarget ~ exits   ~ setentry( newL, newL )
            END

            ITERATE                             /* read next line */
         END
      END

      uLine = TRANSLATE(newL)                   /* rest of directive in uppercase       */

      IF      directive = "METHOD"   THEN       /* -------------------------------------------- */
      DO
         attribute = (WORDPOS("ATTRIBUTE" , uLine) > 0)
         private   = (WORDPOS("PRIVATE"   , uLine) > 0)
         unguarded = (WORDPOS("UNGUARDED" , uLine) > 0)
         protected = (WORDPOS("PROTECTED" , uLine) > 0)
         class     = (WORDPOS("CLASS"     , uLine) > 0)

         IF tmpClass = .nil THEN              /* "floating" method, not attached yet */
            class = .nil

         /* remove quotes, if there */
         name = strip_quotes( name )

/*
say "---> method, name" pp( name ) "class" pp( class )
*/

         tmpObject = .def_method ~ new(name, attribute, private, unguarded, protected, class)
         tmpObject ~ LineNr = line?         /* store line # of definition */
         tmpObject ~ Errors = tmpErrors     /* assign temporary error-list */


         tmpTarget = tmpClass
         errString = ""

         IF class = .true THEN
         DO
            IF tmpClass ~ local_Class_Methods ~ entry( name ) = .nil THEN
            DO
               tmpClass ~ local_Class_Methods ~ setentry( name, tmpObject )
            END
            ELSE
            DO
               errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                    pp( "CLASS METHOD" ) "- already defined for this" tmpTarget ~ type "!"
            END

            ctl.eClassMethods2Class[ tmpClass ] = tmpObject /* relate class to class method */
         END
         ELSE IF class = .false THEN
         DO
            IF tmpClass ~ local_Instance_Methods ~ entry( name ) = .nil THEN
            DO
               tmpClass ~ local_Instance_Methods ~ setentry( name, tmpObject )
            END
            ELSE
            DO
               errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                    pp( "INSTANCE METHOD" ) "- already defined for this" tmpTarget ~ type "!"
            END

            ctl.eInstanceMethods2Class[ tmpClass ] = tmpObject /* relate class to instance method */
         END
         ELSE                   /* floating method (not attached to any class */
         DO
            tmpTarget = .nil            /* no target to attach to */
            IF tmpFile ~ local_Methods ~ entry( name ) = .nil THEN
            DO
               tmpFile ~ local_Methods ~ setentry( name, tmpObject )
            END
            ELSE
            DO
               errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                    pp( "METHOD" ) "- already defined for this" tmpFile ~ type "!"
            END
         END

         IF errString <> "" THEN                /* did duplicate error occur ? */
         DO
            CALL sayError  errString
            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
            tmpObject ~ Errors ~ insert( errString ) /* store error with duplicate object ! */
            IF tmpTarget <> .nil THEN
               tmpTarget ~ Errors ~ insert( errString ) /* store error with target class ! */
         END

/*
         ctl.eMethods ~ setentry( name, tmpObject)      /* add to methods repository */
*/

         ctl.eToken2Files[ tmpFile ] = tmpObject    /* add to relation */
         ctl.eMethods2Files[ tmpFile ]   = tmpObject    /* add to relation */

         ctl.eMethods2Class[  tmpClass ] = tmpObject    /* relate class to method */

         tmpMethod = tmpObject                       /* remember method (for args, returns, ...) */
         noRequires = .true                          /* ::REQUIRES not allowed anymore */


/*** ???? not desired at the moment **** */
         IF attribute THEN
         DO
            tmpMethod ~ expose ~ setentry( name, name )  /* attribute method implicitly exposes itself */
            IF tmpClass <> .nil THEN
            DO
               IF class = .true THEN      /* belongs to class-scope */
               DO
                  tmpClass ~ ExposeClass ~ setentry( name, name )
               END
               ELSE                       /* belongs to instance-scope */
               DO
                  tmpClass ~ ExposeInstance ~ setentry( name, name )
               END
            END
         END
/* *** ???? not desired at the moment ****/


/* ?? check whether floating method has attributes like CLASS, ATTRIBUTE etc. ... */
/* assuming EXPOSE not allowed for a floating method ? REALLY ????? */

         IF tmpClass <> .nil THEN               /* presently class defined ? */
         DO
            /* now check whether next line EXPOSEs object variables ! */
            i = i + 1
            IF i > newArray ~ items THEN ITERATE           /* at end already ? */

            line  = newArray[i, 1]              /* get line */
            line? = newArray[i, 2]              /* get line# in original file */

            PARSE VAR line firstWord newL       /* get expose args */
            firstWord = TRANSLATE( firstWord )

            IF firstWord = "EXPOSE" THEN        /* Object Variable in hand ? */
            DO
               tmpDir = parse_expose( newL )    /* parse expose list into a directory */
               tmpSupp = tmpDir ~ supplier
               DO WHILE tmpSupp ~ available     /* iterate over tmpDir, set method and class */
                  word1 = tmpSupp ~ index
                  word2 = tmpSupp ~ item

                  tmpMethod ~ expose ~ setentry( word1, word2 )  /* store with method */

                  IF class = .true THEN                        /* belongs to class-scope */
                     tmpClass ~ ExposeClass    ~ setentry( word1, word2 )
                  ELSE                                         /* belongs to instance-scope */
                     tmpClass ~ ExposeInstance ~ setentry( word1, word2 )

                  tmpSupp ~ next
               END
            END /* firstWord */
            ELSE
            DO
               i = i - 1                           /* not an expose, reset */

            END
         END
      END



      ELSE IF directive = "ROUTINE"  THEN       /* -------------------------------------------- */
      DO
/***********>
<***********/
         tmpClass = .nil                        /* reset class object */
         public = (WORDPOS("PUBLIC"   , uLine) > 0)
         /* remove quotes, if there */
         name = strip_quotes( name )

         tmpObject = .def_routine ~ new( name, public )
         tmpObject ~ LineNr = line?         /* store line # of definition */
         tmpObject ~ Errors = tmpErrors     /* assign temporary error-list */



         /* routine already defined for target object ? */
         IF tmpFile ~ local_Routines ~ entry( name ) <> .nil THEN
         DO
            errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                 pp( "ROUTINE" ) "- already defined for this" tmpFile ~ type "!"
            CALL sayError  errString
            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
            tmpObject ~ Errors ~ insert( errString ) /* store error with duplicate object ! */
         END
         ELSE
         DO
            tmpFile ~ local_Routines ~ setentry( name, tmpObject )      /* store routine with file */

            IF public THEN                                              /* store public accessible routines */
               tmpFile ~ Visible_Routines ~ setentry( name, tmpObject )
         END


/* ------------->
         IF called_from_file <> .nil THEN               /* add to directory only if public */
         DO
            IF public THEN
               ctl.eRoutines ~ setentry( name, tmpObject)     /* add to routines repository */
         END
         ELSE   /* this is the main file, all routines & classes defined there are accessible */
            ctl.eRoutines ~ setentry( name, tmpObject)  /* add to routines repository */
<-------------- */


         ctl.eToken2Files[ tmpFile ] = tmpObject        /* add to relation */
         ctl.eRoutines2Files[ tmpFile ] = tmpObject     /* add to relation */

         tmpRoutine = tmpObject                         /* remember routine-object */
         noRequires = .true                          /* ::REQUIRES not allowed anymore */
      END





      ELSE IF directive = "CLASS"    THEN       /* -------------------------------------------- */
      DO
/***********>
<***********/

         PARSE VAR line "::" newL               /* remove directive symbols */
         hintline = SUBWORD( workFile[line?], 1, 5) || "..."    /* hint line for error messages */

         /* all class names are either symbols **or** literal strings */
         /* check whether class name is enclosed in quotes */

/*
say
say "CLASS - working on" pp( newL )
*/
         /* look for keyword, return class-name if there */
         res. = get_class_name( "CLASS", newL, line?, hintline, tmpFile )

         IF res.eResult THEN                    /* class name could be found ! */
         DO
            name = res.eName
            newL = res.eNewL                    /* rest of line to be parsed */
         END
         ELSE                                   /* class name could *not* be found */
         DO
            SELECT
               WHEN res.eError  = 1 THEN ITERATE/* string not closed ! */

               WHEN res.eError  = 2 THEN ITERATE/* keyword found, but no class-name left */

               OTHERWISE NOP                    /* keyword not found, no changes, impossible for CLASS-keyword */
            END
         END
/*
say "----> class:" pp( name ) "left" pp( newL )
*/

         /* ------------------------------------------------ */
         /* look for keyword, return class-name if there */
         res. = get_class_name( "SUBCLASS", newL, line?, hintline, tmpFile )


         IF res.eResult THEN                    /* class name could be found ! */
         DO
            subclass = res.eName
            newL = res.eNewL                    /* rest of line to be parsed */
         END
         ELSE                                   /* class name could *not* be found */
         DO
            subclass = .nil                     /* default, if keyword was not found */
            IF res.eError <> .nil THEN
            DO
               newL = res.eNewL                 /* assign what's left for analyzing */
               IF res.eError  = 1 THEN          /* string not closed ! */
                  subclass = "(error: string for class name not closed !)"
               ELSE                             /* keyword found, but no class name left ! */
                  subclass = "(error: class name missing !)"

               tmpErrors ~ insert( res.eErrorString )          /* store with error-list */
            END
         END
/*
say "----> subclass:" pp( subclass ) "left" pp( newL )
*/

         /* ------------------------------------------------ */
         /* look for keyword, return class-name if there */
         res. = get_class_name( "MIXINCLASS", newL, line?, hintline, tmpFile )

         IF res.eResult THEN                    /* class name could be found ! */
         DO
            mixinclass = res.eName
            newL = res.eNewL                    /* rest of line to be parsed */
         END
         ELSE                                   /* class name could *not* be found */
         DO
            mixinclass = .nil                     /* default, if keyword was not found */
            IF res.eError <> .nil THEN
            DO
               newL = res.eNewL                 /* assign what's left for analyzing */
               IF res.eError  = 1 THEN          /* string not closed ! */
                  mixinclass = "(error: string for class name not closed !)"
               ELSE                             /* keyword found, but no class name left ! */
                  mixinclass = "(error: class name missing !)"

               tmpErrors ~ insert( res.eErrorString )          /* store with error-list */
            END
         END
/*
say "----> mixinclass:" pp( mixinclass ) "left" pp( newL )
*/

         /* ------------------------------------------------ */
         IF subclass <> .nil & mixinclass <> .nil THEN
         DO
            errString = pp_lineNr( line?, wfLength ) PP( hintline ) "- SUBCLASS *and* MIXINCLASS defined !"
            CALL sayError  errString
            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
            tmpErrors ~ insert( errString )          /* store with error-list */
         END

         /* ------------------------------------------------ */
         /* look for keyword, return class-name if there */
         res. = get_class_name( "METACLASS", newL, line?, hintline, tmpFile )

         IF res.eResult THEN                    /* class name could be found ! */
         DO
            metaclass = res.eName
            newL = res.eNewL                    /* rest of line to be parsed */
         END
         ELSE                                   /* class name could *not* be found */
         DO
            metaclass = .nil                     /* default, if keyword was not found */
            IF res.eError <> .nil THEN
            DO
               newL = res.eNewL                 /* assign what's left for analyzing */
               IF res.eError  = 1 THEN          /* string not closed ! */
                  metaclass = "(error: string for class name not closed !)"
               ELSE                             /* keyword found, but no class name left ! */
                  metaclass = "(error: class name missing !)"
            END
         END
/*
say "----> metaclass:" pp( metaclass ) "left" pp( newL )
*/

         /* ------------------------------------------------ */
         /* look for keyword, return class-name if there */
         res. = get_class_name( "EXTERNAL", newL, line?, hintline, tmpFile )

         IF res.eResult THEN                    /* class name could be found ! */
         DO
            external = res.eName
            newL = res.eNewL                    /* rest of line to be parsed */
         END
         ELSE                                   /* class name could *not* be found */
         DO
            external = .nil                     /* default, if keyword was not found */
            IF res.eError <> .nil THEN
            DO
               newL = res.eNewL                 /* assign what's left for analyzing */
               IF res.eError  = 1 THEN          /* string not closed ! */
                  external = "(error: external string not closed !)"
               ELSE                             /* keyword found, but no class name left ! */
                  external = "(error: external string missing !)"

               tmpErrors ~ insert( res.eErrorString )          /* store with error-list */
            END
         END
/*
say "----> external:" pp( external ) "left" pp( newL )
*/

         /* ------------------------------------------------ */


         /* ------------------------------------------------------ */
         /* attention, the following two statements *must* be last */
         /* ------------------------------------------------------ */
         uLine = TRANSLATE( newL )                              /* put remaining string into uppercase */

         public = ( WORD( uLine, 1 ) = "PUBLIC" )               /* PUBLIC directive ? */

         pos = WORDPOS("INHERIT",  uLine)
         IF pos = 0 THEN inherit  = .nil
         ELSE
         DO
            IF token_in_string( newL, pos, line? ) THEN         /* error "INHERIT" under quotes ! */
            DO
               errString = pp_lineNr( line?, wfLength ) PP( hintline ) pp( "INHERIT" ) "- INHERIT keyword is under quotes ! "
               CALL sayError  errString
               tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
               tmpErrors ~ insert( errString )          /* store with error-list */

            END
            ELSE
            DO
               rest = SUBWORD( newL, pos + 1 )          /* get list of classes to inherit */
               inheritArray = parse_tokens( rest )          /* parse inherit tokens */

               IF inheritArray ~ items = 0 THEN             /* no list given ! */
               DO
                  errString = pp_lineNr( line?, wfLength ) PP( hintline ) pp( "INHERIT" ) "- INHERIT list is empty !"
                  CALL sayError  errString
                  tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
                  tmpErrors ~ insert( errString )          /* store with error-list */
               END
               ELSE
               DO
                  tmpInheritDir = .directory ~ new              /* directory to check for duplicates */

                  inherit = .list ~ new                         /* create an empty list for inheritance purposes */
                  DO j = 1 TO inheritArray ~ items / inheritArray ~ dimension
                     IF inheritArray[ j, 2 ] = .nil THEN        /* error, string without closing  quotes ! */
                     DO
                        errString = pp_lineNr( line?, wfLength ) PP( hintline ) pp( "INHERIT" ),
                                    "- ending quote missing on" pp( inheritArray[ j, 1 ] )

                        CALL sayError  errString
                        tmpFile ~ errors ~ insert( errString )  /* store error with file-object */
                        tmpErrors ~ insert( errString )         /* store with error-list */
                     END
                     ELSE                                       /* save token */
                     DO
                        tmpInheritString = inheritArray[ j, 1 ]
                        IF \ tmpInheritDir ~ HASENTRY( tmpInheritString ) THEN /* class not processed yet */
                        DO
                           inherit ~ insert( tmpInheritString )
                           tmpInheritDir ~ setentry( tmpInheritString, tmpInheritClass )
                        END
                        ELSE                                    /* error duplicate class found */
                        DO
                           errString = pp_lineNr( line?, wfLength ) PP( hintline ) pp( "INHERIT" ),
                                       "- duplicate class found:" pp( tmpInheritString )

                           CALL sayError  errString
                           tmpFile ~ errors ~ insert( errString )  /* store error with file-object */
                           tmpErrors ~ insert( errString )         /* store with error-list */
                        END
                     END
                  END
               END
            END
         END

/*
say "----> inherit:" pp( inherit ) "left" pp( newL ) "tmpInheritDir" pp( tmpInheritDir )
*/
/*
say "---> after INHERIT, left newL" pp( newL )
*/










         /* ------------------------------------------------------ */

         tmpObject = .def_class ~ new( name, public, subclass, mixinclass,  metaclass, external, inherit )
         tmpObject ~ LineNr = line?         /* store line # of definition */
         tmpObject ~ Errors = tmpErrors     /* assign temporary error-string */


         /* class already defined for target object ? */
         IF tmpFile ~ local_Classes ~ entry( name ) <> .nil THEN
         DO
            errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ),
                 pp( "CLASS" ) "- already defined for this" tmpFile ~ type "!"
            CALL sayError  errString
            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
            tmpObject ~ Errors ~ insert( errString ) /* store error with duplicate object ! */
         END
         ELSE
         DO
            tmpFile ~ local_Classes ~ setentry( name, tmpObject )    /* store routine with file */

            IF public THEN                                              /* store public accessible routines */
               tmpFile ~ Visible_Classes ~ setentry( name, tmpObject )
         END

/* ----------->
         IF called_from_file <> .nil THEN               /* add to directory only if public */
         DO
            IF public THEN
               ctl.eClasses ~ setentry( name, tmpObject) /* add to class  repository */
         END
         ELSE   /* this is the main file, all routines & classes defined there are accessible */
            ctl.eClasses ~ setentry( name, tmpObject)    /* add to class  repository */
<----------- */


         ctl.eToken2Files[ tmpFile ] = tmpObject        /* add to relation */
         ctl.eClasses2Files[ tmpFile ]   = tmpObject    /* add to relation */

         tmpClass  = tmpObject                          /* store class */

         noRequires = .true                             /* ::REQUIRES not allowed anymore */
      END






      ELSE IF directive = "REQUIRES" THEN       /* -------------------------------------------- */
      DO
         IF noRequires THEN                  /* ::REQUIRES *not* allowed anymore ! */
         DO
            errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ) "- *not* allowed anymore ! "
            CALL sayError  errString
            tmpFile ~ errors ~ insert( errString )   /* store error with file-object */

            ITERATE
         END

         name = strip_quotes( name )            /* 97-09-28; strip possible quotes from filename        */

         tmpObject = .def_requires ~ new( name )

         tmpObject ~ LineNr = line?         /* store line # of definition */

         ctl.eToken2Files[ tmpFile ] = tmpObject    /* add to relation */
         ctl.eRequires2Files[ tmpFile ] = tmpObject     /* add to relation */

         IF .local ~ resolve THEN                       /* yes, resolve ::REQUIRES statement */
         DO
            indent = .rgf.util ~ indent
            .rgf.util ~ indent = indent "..."

            call sayError
            call sayError ">>>  recurse, file" pp(name)
            tmpWorkedOnFile = analyze_file( name, tmpFile ) /* <--------- call recursively, i.e. include - REQUIRES-file */
            .rgf.util ~ indent = indent

            CALL sayError "<<<  returned to analyzing" pp(tmpFile ~ name) "..."
            CALL sayError

            /* update present VisibleClasses and VisibleRoutines */
            tmpFile ~ Visible_Classes  = tmpFile ~ Visible_Classes  ~ union( tmpWorkedOnFile ~ Visible_Classes )
            tmpFile ~ Visible_Routines = tmpFile ~ Visible_Routines ~ union( tmpWorkedOnFile ~ Visible_Routines )
         END
      END
      ELSE                                   /* unknown directive */
      DO
         errString = pp_lineNr( line?, wfLength ) PP( SUBWORD( workFile[line?], 1, 5) || "..." ) "- unknown directive !"
         CALL sayError  errString
         tmpFile ~ errors ~ insert( errString )   /* store error with file-object */
         ITERATE
      END
   END

   /* ok, fix up class pointers and mark metaclasses */

   CALL fix_up_classes tmpfile, wfLength

   RETURN tmpFile











/* ------------------------------------------------------------------------------------- */
/*
        make sure that a string of words which is delimited by commas is being delimited
        by commas *and* one blank;
        removes leading and trailing commas !
*/
PP_ARG_LIST: PROCEDURE
   PARSE ARG string

   tmp = ""
   DO WHILE string <> ""
      PARSE VAR string work "," string
      tmp = STRIP( tmp ) || "," work
   END

   tmp = STRIP( tmp, "Both", "," )      /* strip commas */
   tmp = SPACE( tmp )                   /* make sure only one space between words ! */

   RETURN tmp


/* ------------------------------------------------------------------------------------- */
/*
        parse an EXPOSE string (blank delimited and bracketed lists), return a
        directory (serves to remove duplicates)
*/
PARSE_EXPOSE: PROCEDURE
    PARSE ARG newL

    tmpDir = .directory ~ new

    DO WHILE newL <> ""
       PARSE VAR newL var "(" list ")" newL  /* take care of list in expose */

       DO WHILE var <> ""
          PARSE VAR var word var

          tmpDir ~ setentry( word, word )  /* store with method */
       END

       IF list <> "" THEN
       DO
          word = "(" || SPACE( list ) || ")"

          tmpDir ~ setentry( word, word )  /* store with method */
       END
    END
    RETURN tmpDir


/* ------------------------------------------------------------------------------------- */
/* parse given keyword-instruction, if not present return .nil */
PARSE_KEYWORD : PROCEDURE
  PARSE UPPER ARG keyword_Needle, tmpLine

  IF POS( keyword_Needle, tmpLine ) = 0 THEN    /* keyword not found, return immediately */
     RETURN .nil

  tmpDelimiter = "=[(~"         /* if these follow the keyword immediately, then we have
                                   in effect a variable or method name in hand */
  keywordLength = LENGTH( keyword_Needle )

  PARSE ARG ., line             /* get original line */

  expression = ""

  PARSE UPPER VAR line before ( keyword_Needle ) after

  IF before = "" & after = "" THEN
     RETURN ""                /* skip line, no keyword expression in hand */

  /* check, whether keyword is within string */
  strPos = POS( keyword_Needle, tmpLine )

  DO WHILE strPos > 0         /* if 1, then line starts with keyword, i.e. o.k. */
     /* check whether within string */
     IF \token_in_string( line, strPos, line?) THEN   /* extract "before" and "after" */
     DO
        IF strPos < 2 THEN      /* not at beginning */
        DO
           PARSE VAR line keyword after
           before = ""
        END
        ELSE                    /* parse keyword up to next blank */
        DO
           step = strPos - 1
           PARSE VAR line 1 before +(step) keyword after
        END

        /* error, keyword immediately followed by delimiter ! */
        IF POS( tmpDelimiter, SUBSTR( keyword, keywordLength, 1 ) ) > 0 THEN
        DO
           strPos = 0
           LEAVE
        END

        IF after = "" THEN LEAVE

        after = STRIP( after )

        IF before <> "" THEN
           before = TRANSLATE( WORD( before, WORDS( before )) )

        IF before = "THEN" | before = "ELSE" | before = "" THEN   /* found a true keyword ! */
        DO
           tmpChar = LEFT( after, 1 )

           IF POS( tmpChar, tmpDelimiter ) = 0 THEN     /* found the keyword, this time followed by an expression ! */
              LEAVE
        END
     END

     strPos = POS( keyword_Needle, tmpLine, strPos + 1 )    /* look for another "RETURN" */
  END

  IF strPos = 0 THEN RETURN .nil                /* keyword needle not found */

  RETURN after                                  /* return expression (maybe empty string) */






/* ------------------------------------------------------------------------------------- */
/* parse first word, if symbol, fine, else take care of quotes, return a stem containing results:

   returns a stem:

        res.eName        ... class name (i.e. value going with keyword) or .nil if error
        res.eNewL        ... remaining string (keyword plus value extracted)
        res.eResult      ... .true, if value found, .false else
        res.eError       ... error (1 = quotes not closed, 2 = keyword value not found, .nil else )
        res.eErrorString ... error string to be attached to token


*/
GET_CLASS_NAME: PROCEDURE EXPOSE ctl.
   USE ARG keyword, newL, line?, line_hint, tmpFile

/*
say
say "\\\ get_class_name(), keyword" pp( keyword ) "newL" pp( newL )
*/

   res. = .nil                                  /* default for stem */
/*
   SIGNAL ON SYNTAX                             /* activate handling of signal SYNTAX   */
*/

   tmpString    = ""
   oldPos       = 1

   DO FOREVER
      IF oldPos = 0 THEN                        /* nothing to look for, nothing found   */
      DO
         tmpArray[ 1 ] = ""                     /* indicate, nothing found              */
         tmpString     = newL                   /* return unchanged line                */
         LEAVE
      END

      tmpArray     = get_next_token( newL, oldPos, line? )      /* read a token         */

      tmpToken = tmpArray ~ at( 1 )
      IF tmpToken = "" THEN                     /* nothing there !                      */
         LEAVE

      tmpPos   = tmpArray ~ at( 3 )             /* get next position to analyze         */
/*
say "... looping, oldPos" pp( oldPos ) "tmpPos" pp( tmpPos ) "keyword" pp( keyword )
pause
*/

      IF TRANSLATE( tmpToken ) = keyword THEN   /* keyword found, get next token !      */
      DO
/*
say "... BINGO" pp( tmpToken ) "=" pp( keyword ) "found !"
*/
         IF oldPos > 1 THEN                     /* assemble non-processed tokens        */
            tmpString = SUBSTR( newL, 1, oldPos - 1 )

         oldPos = tmpPos                        /* get next token going with keyword    */

         tmpArray  = get_next_token( newL, oldPos, line? )
/*
say "... BINGO" pp( tmpToken ) "going with token" pp( tmpArray ~ at( 1 ) )
*/

         IF tmpArray ~ at( 1 ) = "" THEN        /* error, no token following keyword !  */
         DO
             errString = pp_lineNr( line?, wfLength ) PP( line_hint ) PP( keyword ) "- class name not found !"
             CALL sayError errString
             tmpFile ~ errors ~ insert( errString )   /* store error with file-object   */

             res.eErrorString = errString
             res.eError  = 2                   /* keyword value not found !             */
             res.eResult = .false
             res.eNewL   = tmpString
             RETURN res.
         END
         ELSE                                   /* o.k. finish tmpString and leave      */
         DO
            IF tmpArray ~ at( 3 ) > 0 THEN
               tmpString = tmpString SUBSTR( newL, tmpArray ~ at( 3 ) )
         END
         LEAVE
      END

      oldPos = tmpPos                           /* keyword not found so far             */
   END

   /* assign class-name going with keyword, remove quotes */
   tmpName     = tmpArray ~ at( 1 )
   IF tmpName <> "" THEN                        /* only if token given, assign it       */
   DO
      tmpName     = strip_quotes( tmpArray ~ at( 1 ) )
      res.eName   = tmpName
   END

   res.eNewL   = tmpString
   res.eResult = .true
   RETURN res.


SYNTAX :                                        /* error in get_next_token()            */
   errString = pp_lineNr( line?, wfLength ) PP( line_hint ) PP( keyword ) "- quotes for class name not closed !"
   CALL sayError errString
   tmpFile ~ errors ~ insert( errString )       /* store error with file-object */
   res.eNewL        = tmpString
   res.eErrorString = errString
   res.eError       = 1                         /* string not closed ! */
   res.eResult      = .false
   RETURN res.







/* ------------------------------------------------------------------------------------- */
/* parse first word, if symbol, fine, else take care of quotes, return an array, containing found tokens:

        resArray[ n, 1 ] ... contains token or string in error, if token resides within malformed string (no ending quote)
        resArray[ n, 2 ] ... contains .true, if token was originally embedded within quotes,
                                      .false else (i.e. a symbol ),
                                      .nil   if ending quote is missing !
*/
PARSE_TOKENS: PROCEDURE EXPOSE ctl.
   USE ARG line

   resArray = .array ~ new              /* will contain an array of tokens as a result */
   i        = 1

   IF line = "" THEN RETURN resArray    /* return empty array */


   DO WHILE line <> ""
      line = STRIP( line )
      parseLine = line

      first_Word = WORD( line, 1 )              /* get first word after keyword */
      quote = LEFT( first_Word, 1 )             /* is word starting with a quote ? */

      IF quote = "'" | quote = '"' THEN
      DO
         token_name = ""
         quote_active = .true
         pos = 2                                /* point to char after first quote */
         next_char = SUBSTR( line, pos, 1 )     /* extract next char */

         IF next_char = quote THEN
         DO
            IF SUBSTR( line, 3, 1) <> quote THEN   /* a ""_ found, i.e. nothing */
            DO
               line = SUBSTR( line, 3 )         /* "eat" 'nothing' */
               ITERATE
            END
         END

         linelength = LENGTH( line )            /* get line length */

         DO WHILE pos > 0 & pos < lineLength

            token_name = token_name || next_char

            tmp_char = next_char                /* store next char */

            pos = pos + 1
            next_char = SUBSTR( line, pos, 1 )  /* get next char */

            IF next_char = quote THEN
            DO
               IF tmp_char = quote THEN         /* ok, a pair of _"_"_ found */
               DO
                  pos = pos + 1
                  IF pos > lineLength THEN quote_active = .false

                  next_char = SUBSTR( line, pos, 1 )
               END
               ELSE                             /* looks like the end of a quote ? */
               DO
                  pos = pos + 1
                  next_char = SUBSTR( line, pos, 1 )
                  IF next_char <> quote THEN    /* Yup, end of the quote */
                  DO
                     quote_active = .false
                     LEAVE
                  END
               END
            END
         END

         IF quote_active THEN                   /* error, quotes not closed ! */
         DO
            resArray[ i, 1 ] = parseLine        /* store the string with missing quote */
            resArray[ i, 2 ] = .nil             /* indicate that a quote was missing  */
            LEAVE
         END
         ELSE
         DO
            line = STRIP( SUBSTR( line, pos + 1) ) /* remaining tokens */
            resArray[ i, 1 ] = token_name       /* store token                        */
            resArray[ i, 2 ] = .true            /* indicate that a quote was active   */
         END
      END
      ELSE                                      /* just extract word */
      DO
         resArray[ i, 1 ] = first_Word          /* store token                        */
         resArray[ i, 2 ] = .false              /* indicate that a quote was *not* active   */
         PARSE VAR line . line
      END
      i = i + 1
   END

   RETURN resArray




/* ------------------------------------------------------------------------------------- */

/* return the line number in edited form */
PP_LINENR : PROCEDURE
   PARSE ARG text, width

   IF text = "" THEN RETURN ""                  /* no line # given */

   IF DATATYPE( width, "W" ) THEN               /* width given */
      RETURN "@ l#" pp( RIGHT( text, width ) ) "-"
   ELSE                                         /* no width available */
      RETURN "@ l#" pp( text ) "-"



/* ------------------------------------------------------------------------------------- */

::REQUIRES RGF_UTIL.REX
::REQUIRES ORX_UTIL.REX

/*
::REQUIRES "T:\orx\test\test_cls_res\cltree.cmd"
*/


/* ------------------------------------------------------------------------------------- */



/* ------------------------------------------------------------------------------------- */


/* find file according to PATH-statement, starting out with present directory */
::ROUTINE find_file2
  USE ARG file_name, path

  DO 2
     tmpPath = STREAM( file_name, "C", "QUERY EXISTS" )    /* 97-09-28 */
     IF tmpPath <> "" THEN RETURN tmpPath                  /* 97-09-28 */

     IF \ VAR( "path" ) THEN       /* get path     */
        path = ".;" || VALUE( .analyze.local ~ path, , .analyze.local ~ selector )

     tmpFile = FIND_FILE( strip_quotes( file_name ), path )   /* use external routine */
     IF tmpFile <> .nil THEN RETURN tmpFile

     /* not found, now try all uppercased file-name     */
     file_name = translate( file_name ) /* ??? 99-01-09, 'cause 'ix need *exact* case */
  END

  SIGNAL ON SYNTAX
  RAISE SYNTAX 3.901 ARRAY (file_name)  /* file not found, raise syntax error   */

SYNTAX:
  RAISE PROPAGATE                       /* raise syntax-error in caller */




/* accept a ClassID or a Def_Class and search for it in the environment */
:: ROUTINE SEARCH_ENVIRONMENT
   USE ARG ClassNameString

   IF \ IsA( ClassNameString, .string ) THEN    /* a def_class in hand ? */
      ClassNameString = ClassNameString ~ name

   RETURN .Env.FileObj ~ Local_Classes ~ entry( ClassNameString  )



/****************************************************************************************/

/* ************************************************************************* */
/* purpose: define those methods which are used by almost all subclasses */
::CLASS def
/* -------------------------- */
::METHOD INIT                   CLASS
  EXPOSE max_name_length total_objects

  max_name_length = 0
  total_objects   = 0
  self ~ User_Slot      = .directory ~ new      /* a directory for getting user-infos from other programs */

  FORWARD CLASS (super)

:: METHOD max_name_length       CLASS   ATTRIBUTE
:: METHOD total_objects         CLASS   ATTRIBUTE
:: METHOD User_Slot             CLASS   ATTRIBUTE       /* if a user wants to add individual data */
/* -------------------------- */

::METHOD INIT
  EXPOSE  name
  USE ARG name

  self ~ Errors         = .list ~ new
  self ~ Exits          = .directory ~ new
  self ~ IsFunction     = .false
  self ~ IsProcedure    = .false
  self ~ LineNr         = .nil
  self ~ Local_Labels     = .directory ~ new    /* defined in this file ... */
  self ~ Local_Procedures = .directory ~ new    /* defined in this file ... */
  self ~ Returns        = .directory ~ new
  self ~ Signatures     = .directory ~ new
  self ~ User_Slot      = .directory ~ new      /* a directory for getting user-infos from other programs */


  /* store largest name with class */
  self ~ class ~ max_name_length = MAX( self ~ class ~ max_name_length, LENGTH( name ) )
  self ~ class ~ total_objects = self ~ class ~ total_objects + 1

:: METHOD Errors                ATTRIBUTE       /* errors found on parsing file */
:: METHOD Exits                 ATTRIBUTE       /* return statement             */
:: METHOD IsFunction            ATTRIBUTE       /* is this a function ?         */
:: METHOD IsProcedure           ATTRIBUTE       /* is this a procedure?         */
:: METHOD LineNr                ATTRIBUTE       /* line # of token              */
:: METHOD Local_Labels          ATTRIBUTE       /* labels for this token        */
:: METHOD Local_Procedures      ATTRIBUTE       /* procedures for this token */
:: METHOD Name                  ATTRIBUTE       /* file's name                  */
:: METHOD Returns               ATTRIBUTE       /* return statement             */
:: METHOD Signatures            ATTRIBUTE       /* USE ARG, PARSE ARG           */
:: METHOD Type                                  /* to be defined by subclasses */
:: METHOD User_Slot             ATTRIBUTE       /* if a user wants to add individual data */

:: METHOD DumpAttributes/* to be defined by subcalsses */
   tmp = ""
   RETURN tmp



/* ************************************************************************* */
::CLASS def_file        SUBCLASS def
/* -------------------------- */
::METHOD INIT

  FORWARD CLASS ( super ) CONTINUE              /* let super initialize first */

  self ~ Loc                   = 0
  self ~ Local_Classes         = .directory ~ new    /* defined in this file ... */
  self ~ Local_Leaf_Classes    = .set       ~ new    /* defined in this file ... */
  self ~ Local_Metaclasses     = .directory ~ new    /* stores metaclasses ....  */
  self ~ Local_Methods         = .directory ~ new    /* defined in this file ... */
  self ~ Local_Root_Classes    = .set       ~ new    /* defined in this file ... */
  self ~ Local_Routines        = .directory ~ new    /* defined in this file ... */
  self ~ Required_by           = .list ~ new
  self ~ Requires_files        = .list ~ new
  self ~ ShortName             = FILESPEC( "Name", self ~ name )        /* extract path */
  self ~ Total_loc             = 0
  self ~ Visible_Classes       = .directory ~ new    /* according to scope rules ... */
  self ~ Visible_Routines      = .directory ~ new    /* according to scope rules ... */

:: METHOD LOC                   ATTRIBUTE       /* lines of code (stripped)     */
:: METHOD Local_Classes         ATTRIBUTE       /* locally defined classes      */
:: METHOD Local_Leaf_Classes    ATTRIBUTE       /* locally defined leaf-classes (a Set) */
:: METHOD Local_Metaclasses     ATTRIBUTE       /* directory of local metaclasses */
:: METHOD Local_Methods         ATTRIBUTE       /* methods for file itself      */
:: METHOD Local_Root_Classes    ATTRIBUTE       /* locally defined root-classes (a Set) */
:: METHOD Local_Routines        ATTRIBUTE       /* locally defined routines     */
:: METHOD Required_by           ATTRIBUTE       /* files, which require this one */
:: METHOD Requires_files        ATTRIBUTE       /* files required ...           */
:: METHOD ShortName             ATTRIBUTE       /* short file name (w/o path)   */
:: METHOD Total_loc             ATTRIBUTE       /* lines of code                */
:: METHOD Visible_Classes       ATTRIBUTE       /* total classes visible        */
:: METHOD Visible_Routines      ATTRIBUTE       /* total routines visible       */

::METHOD type
  RETURN "FILE"


/* ************************************************************************* */
::CLASS def_procedure   SUBCLASS def
/* -------------------------- */
::METHOD INIT
  EXPOSE procedure
  USE ARG name , procedure

  FORWARD CLASS ( super ) CONTINUE      /* let super initialize first */

  self ~ Expose = .directory ~ new

::METHOD Expose         ATTRIBUTE       /* EXPOSE on PROCEDURE line */

::METHOD Type
  RETURN "PROCEDURE"

::METHOD dumpAttributes
  EXPOSE expose
  tmp = ""

  tmpArray = sort( expose )             /* sort directory */
  DO i = 1 TO tmpArray ~ items         /* build sorted EXPOSE list */
     tmp = tmp expose ~ entry( tmpArray[ i ] )
  END

  IF tmp = "" THEN RETURN tmp
              ELSE RETURN "EXPOSE" STRIP( STRIP( tmp, , "," ) )


/* ************************************************************************* */
::CLASS def_label SUBCLASS def_procedure

::METHOD type
  RETURN "LABEL"
/* -------------------------- */




/* ************************************************************************* */
::CLASS def_routine     SUBCLASS def
::METHOD INIT
  EXPOSE public

  USE ARG name, public

  FORWARD CLASS ( super )               /* let super initialize first */


::METHOD public         ATTRIBUTE       /* .true, if public routine    */

::METHOD type
  RETURN "::ROUTINE"

::METHOD dumpAttributes

  tmp = ""

  IF self ~ public THEN tmp = "PUBLIC"
                   ELSE tmp = "      "
  RETURN tmp

/* ************************************************************************* */
::CLASS def_requires    SUBCLASS def

::METHOD INIT

  FORWARD CLASS ( super )               /* let super initialize first */

::METHOD type
  RETURN "::REQUIRES"


/* ************************************************************************* */
::CLASS def_method      SUBCLASS def

::METHOD INIT
  EXPOSE attribute private unguarded protected class_method
  USE ARG name, attribute, private, unguarded, protected, class_method

  self ~ Expose           = .directory ~ new

  IF name = "" THEN                      /* abbutal or blank method ?    */
  DO
     IF name == "" THEN                  /* abbutal ! */
        name = "'' (hint: 'abbutal'-method, i.e. null string !)"
     ELSE
        name = "' ' (hint: 'blank'-method !)"
  END

  FORWARD CLASS ( super ) ARRAY ( name )/* let super initialize */


::METHOD Attribute      ATTRIBUTE       /* .true if attribute           */
::METHOD Class_method   ATTRIBUTE       /* .true if class_method method, .false if instance method,
                                           .nil if non-class_method method    */
::METHOD Expose         ATTRIBUTE       /* EXPOSE string                */
::METHOD Private        ATTRIBUTE       /* .true if private             */
::METHOD Protected      ATTRIBUTE       /* .true if protected           */
::METHOD Unguarded      ATTRIBUTE       /* .true if not guarded         */


::METHOD type
  RETURN "::METHOD"

::METHOD dumpAttributes
  USE ARG bWithSpaces

  IF \VAR( "bWithSpaces" ) THEN bWithSpaces = .true
  tmp = ""

  IF  self ~  class_method = .nil THEN NOP          /* tmp = " --- " */
  ELSE IF self ~ class_method     THEN tmp = "class"
                                  ELSE IF bWithSpaces THEN tmp = "     "

  IF  self ~  attribute    THEN tmp = tmp "attribute"
                           ELSE IF bWithSpaces THEN tmp = tmp "         "

  IF  self ~  private      THEN tmp = tmp "private"
                           ELSE IF bWithSpaces THEN tmp = tmp "       "

  IF  self ~  unguarded    THEN tmp = tmp "unguarded"
                           ELSE IF bWithSpaces THEN tmp = tmp "         "

  IF  self ~  protected    THEN tmp = tmp "protected"
                           ELSE IF bWithSpaces THEN tmp = tmp "         "
  RETURN STRIP( tmp )

:: METHOD ExposeAsString                /* return EXPOSE-directory as string */
   EXPOSE expose

  tmp = ""

  tmpArray = sort( expose )             /* sort directory */
  DO i = 1 TO tmpArray ~ items          /* build sorted EXPOSE list */
     tmp = tmp expose ~ entry( tmpArray[ i ] )
  END

  IF tmp <> "" THEN RETURN "EXPOSE" STRIP( STRIP( tmp, , "," ) )

  RETURN tmp




/* ************************************************************************* */
::CLASS def_class       SUBCLASS def
/* -------------------------- */
::METHOD INIT
  EXPOSE  public subclass mixinclass  metaclass external inherit
  USE ARG name, public, subclass, mixinclass,  metaclass, external, inherit

  FORWARD CLASS ( super ) CONTINUE      /* let super initialize             */

/*
  self ~ External    = external
  self ~ Inherit     = inherit
  self ~ Metaclass   = metaclass
  self ~ Mixinclass  = mixinclass
  self ~ Public      = public
  self ~ Subclass    = subclass
*/

  self ~ ExposeClass            = .directory ~ new    /* class variables          */
  self ~ ExposeInstance         = .directory ~ new    /* instance variables       */
  self ~ IsMetaClass            = .false              /* default to an object class */
  self ~ IsUsedAsSuper          = .false              /* default to a leaf class  */
  self ~ ListOfSuperClasses     = .list ~ new         /* list of superclasses ... */
  self ~ Local_Class_Methods    = .directory ~ new    /* defined in this file ... */
  self ~ Local_Instance_Methods = .directory ~ new    /* defined in this file ... */
  self ~ MetaClassObject        = .nil                /* default to missing class ... */
  self ~ MetaUsedBySet          = .set       ~ new    /* stores classes using this class as a metaclass */
  self ~ SetOfSubclasses        = .set ~ new          /* set of classes, subclassing this class */
  self ~ SuperClassObject       = .nil                /* default to missing class ... */

:: METHOD ExposeClass             ATTRIBUTE     /* class object variables                */
:: METHOD ExposeInstance          ATTRIBUTE     /* instance object variables             */
:: METHOD External                ATTRIBUTE     /* external-string                       */
:: METHOD Inherit                 ATTRIBUTE     /* list of classids of inherited classes */
:: METHOD IsMetaClass             ATTRIBUTE     /* descendant of class class (metaclass)?*/
:: METHOD IsUsedAsSuper           ATTRIBUTE     /* is this class used as a superclass ?  */
:: METHOD ListOfSuperClasses      ATTRIBUTE     /* list of superclasses ...              */
:: METHOD Local_Class_Methods     ATTRIBUTE     /* locally defined class methods         */
:: METHOD Local_Instance_Methods  ATTRIBUTE     /* locally defined instance methods      */
:: METHOD MetaClass               ATTRIBUTE     /* class which is metaclass              */
:: METHOD MetaClassObject         ATTRIBUTE     /* def_class of metaclass                */
:: METHOD MetaUsedBySet           ATTRIBUTE     /* set of classes using this class as a meta class */
:: METHOD MixinClass              ATTRIBUTE     /* classid which gets subclassed         */
:: METHOD Public                  ATTRIBUTE     /* .true, if public class                */
:: METHOD SetOfSubclasses         ATTRIBUTE     /* set of classes, subclassing this one  */
:: METHOD SubClass                ATTRIBUTE     /* classid which gets subclassed         */
:: METHOD SuperClassObject        ATTRIBUTE     /* def_class of superclass               */


::METHOD Type
  RETURN "::CLASS"

::METHOD dumpAttributes
  USE ARG bWithSpaces

  IF \VAR( "bWithSpaces" ) THEN bWithSpaces = .true
  tmp = ""

  IF self ~ public = .nil         THEN
                                  DO
                                     IF bWithSpaces THEN
                                        tmp = "       "       /* "PuBlIc?" */
                                  END
  ELSE IF self ~ public           THEN
                                  DO
                                     IF bWithSpaces THEN
                                        tmp =     "PUBLIC "
                                     ELSE
                                        tmp =     "PUBLIC"
                                  END
                                  ELSE IF bWithSpaces THEN tmp =     "       "

  IF self ~ mixinclass    <> .nil THEN tmp = tmp "mixinclass" pp( self~mixinclass )
  ELSE IF self ~ subclass <> .nil THEN
                                  DO /* don't show default .OBJECT */
                                     IF search_environment( self ~ SubClass ) <> .Env.Object THEN
                                     DO
                                        IF bWithSpaces THEN
                                           tmp = tmp "subclass  " pp( self~subclass )
                                        ELSE
                                           tmp = tmp "subclass" pp( self~subclass )
                                     END
                                  END
                                  ELSE IF bWithSpaces THEN tmp = tmp "          "

                                       /* don't default show .CLASS */
  IF self ~ metaclass <> .nil     THEN IF search_environment( self ~ metaclass ) <> .Env.Class THEN
                                          tmp = tmp "metaclass " pp( self~metaclass )

  IF self ~ external <> .nil      THEN tmp = tmp "external" pp( self ~ external )

  IF self ~ inherit  <> .nil      THEN
  DO
     tmp = tmp "inherit"
     tmpObject = self ~ inherit         /* get inheritance list                 */
     DO item OVER tmpObject
        tmp = tmp pp( item )
     END
  END

  RETURN STRIP( tmp )


